module LatinHypercubeSampling

export  randomLHC,
        AudzeEgliasObjective,
        AudzeEgliasObjective!,
        LHCoptim,
        LHCoptim!,
        subLHCoptim,
        subLHCindex,
        Categorical,
        Continous,
        LHCDimension

using StatsBase
using Random
import Random.randperm

abstract type LHCDimension end

struct Categorical <: LHCDimension
    levels::Int64
end

struct Continous <: LHCDimension
end


include("GA.jl")
include("AudzeEgliasObjective.jl")


function randperm(dim::Continous,n)
    randperm(n)
end

function randperm(dim::Categorical,n)
    lvls = dim.levels
    out = Array{Int64}(undef,n)

    for i = 1:lvls
        s = round(Int,(i-1)*n/lvls) + 1
        e = round(Int,i*n/lvls)
        out[s:e] .= i
    end
    shuffle!(out)    
end


"""
    function randomLHC(n::Int, d::Int)
Generate a random Latin Hypercube with `d` dimensions and `n` sample points.
"""
function randomLHC(n::Int,d::Int)

    dims = [Continous() for i in 1:d]
        
    return randomLHC(n,dims)

end


function randomLHC(n::Int,dims::T) where T<:AbstractArray{S} where S<:LHCDimension

    LHC = Array{Int}(undef,n,length(dims))
    for (i, dim) = enumerate(dims)
        LHC[:,i] = randperm(dim,n)
    end

    return LHC

end




"""
    function LHCoptim(n::Int,d::Int,gens;popsize::Int=100,ntour::Int=2,ptour=0.8)
Produce an optimized Latin Hyper Cube with `d` dimensions and `n` sample points.
Optimization is run for `gens` generations.
"""
function LHCoptim(n::Int,d::Int,gens;   popsize::Int=100,
                                        ntour::Int=2,
                                        ptour=0.8,
                                        dims=[Continous() for i in 1:d],
                                        weights=ones(Float64,length(dims)).*1/length(dims))

    #populate first individual
    X = randomLHC(n,d)

    LHCoptim!(X,gens,popsize=popsize,ntour=ntour,ptour=ptour,dims=dims,weights=weights)

end



"""
    function LHCoptim!(X::Array{Int,2},n::Int,d::Int,gens;ntour::Int=2,ptour=0.8)
Same as LHCoptim(n::Int,d::Int,gens;popsize::Int=100,ntour::Int=2,ptour=0.8) but using an
existing population. Useful for continued optimization.
"""
function LHCoptim!(X::Array{Int,2},gens;    popsize=100,
                                            ntour::Int=2,
                                            ptour=0.8,
                                            dims=[Continous() for i in 1:d],
                                            weights=ones(Float64,length(dims)).*1/length(dims))

    #preallocate memory
    n, d = size(X)                                  #Num points, num dimensions
    mut_inds = Array{Int}(undef,2)          #Storage of indices to swap in mutation
    tour_inds = Array{Int}(undef,ntour)     #Storage of indices for tournament selection
    tour_fitinds = Array{Int}(undef,ntour)  #Storage of fitness for tournament selection
    
    #allocate first population
    pop = Array{eltype(X)}(undef,popsize,n,d)     
    pop[1,:,:] = X
    for i = 2:popsize
        pop[i,:,:] = randomLHC(n,dims)
    end

    nextpop = Array{Int}(undef, popsize,n,d)
    fitness = Vector{Float64}(undef, popsize)
    bestfits = Array{Float64}(undef, gens)
    dist = zeros(Float64,Int(n*(n-1)*0.5))

    #crossover for even population size
    popEven = iseven(popsize)*-1

    #dynamic mutation rate table
    muts = Array{Int}(undef, gens)
    for l = 1:gens
        muts[l] = round(Int,-n/(gens*0.75)*l+n)
        if muts[l] < 1
            muts[l] = 1
        end
    end


    #evaluate first populations fitness
    for i = 1:popsize
        fitness[i] = AudzeEgliasObjective!(dist,  view(pop,i,:,:);
                                            weights=weights,dims=dims)
    end


    #save the best individual and its fitness
    bestfit, bestind = findmax(fitness)
    nextpop[1,:,:] = pop[bestind,:,:]
    bestfits[1] = bestfit

    #ensure fixed crossover is only applied to the continous dimensions
    continousDims = findall(dims.==Ref(Continous()))

    #iterate for gens generations
    for k = 1:gens

        #tournament selection
        for i = 2:popsize
            winnerInd = tournament!(fitness,ntour,tour_inds,tour_fitinds,ptour)
            nextpop[i,:,:] = pop[winnerInd,:,:]
        end
        
        #create children from crossover
        for i = 2:2:popsize+popEven
            for j in continousDims
                if rand() < 1.0/length(continousDims)
                    nextpop[i,:,j], nextpop[i+1,:,j] = fixedcross(view(pop,i,:,j), view(pop,i+1,:,j))
                end
            end
        end

        # #perform inversion and mutation
        for i = 2:popsize
            for j = 1:d
                if rand() < 1.0/d     #probability for inversion
                    inversion!(view(nextpop,i,:,j))
                end
            end

            for j=1:muts[gens]
                mutateLHC!(view(nextpop,i,:,:),mut_inds)
            end
        end

        #evaluate fitness
        for i = 1:popsize
            fitness[i] = AudzeEgliasObjective!(dist, view(nextpop,i,:,:);
            weights=weights,dims=dims)
        end

        #set altered population to current
        pop = nextpop

        #set the first individual to the best and save the fitness
        bestfit, bestind = findmax(fitness)
        nextpop[1,:,:] = pop[bestind,:,:]
        bestfits[k] = bestfit
    end

    return nextpop[1,:,:], bestfits

end



"""
    function subLHCoptim(X,n::Int,gens;popsize::Int=100,ntour::Int=2,ptour=0.8)
Produce an optimized Latin Hyper Cube with `n` sample points from a subset of
points in `X`. Optimization is run for `gens` generations.
"""
function subLHCoptim(X,n::Int,gens;popsize::Int=100,ntour::Int=2,ptour=0.8)

    #preallocate memory
    nLarge, d = size(X)
    pop = Array{Int}(undef, popsize+1,n,d)
    nextpop = Array{Int}(undef, popsize+1,n,d)
    fitness = Vector{Float64}(undef, popsize+1)
    bestfits = Array{Float64}(undef, gens)
    dist = zeros(Float64,Int(n*(n-1)*0.5))

    tour_inds = Array{Int}(undef,ntour)     #Storage of indices for tournament selection
    tour_fitinds = Array{Int}(undef,ntour)  #Storage of fitness for tournament selection

    #dynamic mutation rate table
    muts = Array{Int}(undef, gens)
    for l = 1:gens
        muts[l] = round(Int,-n/(gens*0.75)*l+n)
        if muts[l] < 1
            muts[l] = 1
        end
    end


    #populate first population with random subLHC's and evaluate fitness
    for i = 1:popsize+1
        subInds = sample(1:nLarge, n, replace = false)
        pop[i,:,:] = X[subInds,:]
        fitness[i] = AudzeEgliasObjective!(dist, view(pop,i,:,:))
    end


    #save the best individual and it's fitness
    bestfit, bestind = findmax(fitness)
    nextpop[1,:,:] = pop[bestind,:,:]
    bestfits[1] = bestfit


    #iterate for gens generations
    for k = 1:gens
        
        #tournament selection
        for i = 2:popsize+1
            winnerInd = tournament!(fitness,ntour,tour_inds,tour_fitinds,ptour)
            nextpop[i,:,:] = pop[winnerInd,:,:]
        end

        #perform mutation
        for i = 2:popsize+1
            for j=1:muts[gens]
                subPoint = sample(1:n) #Randomly chosen point subLHC
                largePoint = sample(1:nLarge) #Randomly chosen point largeLHC

                while nextpop[i,subPoint,:] == X[largePoint,:]
                    largePoint = sample(1:nLarge) #Choose new random point in largeLHC
                end
                nextpop[i,subPoint,:] = X[largePoint,:]
            end
        end

        #evaluate fitness
        for i = 1:popsize+1
            fitness[i] = AudzeEgliasObjective!(dist, view(nextpop,i,:,:))
        end

        #set altered population to current
        pop = nextpop

        #set the first individual to the best and save the fitness
        bestfit, bestind = findmax(fitness)
        nextpop[1,:,:] = pop[bestind,:,:]
        bestfits[k] = bestfit
    end

    return nextpop[1,:,:], bestfits

end



"""
    function subLHCindex(X,Xsub)
Index in the large LHC to get the subLHC.
"""
function subLHCindex(X,Xsub)

    nsub = size(Xsub,1)
    subInds = Array{Int}(undef, nsub)

    for i = 1:nsub
        A = all(Xsub[i,:]' .== X, dims=2)
        subInds[i] = (LinearIndices(A))[findall(A)][1]
    end
    return subInds
end


end # module
